from typing import Any, List, Optional, Tuple
from pydantic import Field
from langchain.schema import BaseRetriever, Document


class myRerankRetriever(BaseRetriever):
    model_name: str = "BAAI/bge-reranker-base"
    """rerank model to use."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict] = Field(default=None)
    docs: List[Document] = Field(default=None)
    k: int = 20
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None

    @classmethod
    def from_documents(
        cls,
        docs: List[Document],
        k: int = 20,
        relevancy_threshold: float = 0.0,
        model_name: str = "BAAI/bge-reranker-base",
        **kwargs: Any,
    ) -> BaseRetriever:
        return cls(
            model_name=model_name,
            texts=[doc.page_content for doc in docs],
            metadata=[doc.metadata for doc in docs],
            k=k,
            docs=docs,
            relevancy_threshold=relevancy_threshold,
            **kwargs,
        )

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement rerank to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """

        top_k_results, top_k_scores = rerank(
            query, self.docs, self.relevancy_threshold, self.model_name
        )

        return top_k_results, top_k_scores

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    async def _aget_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]


def rerank(query: str, docs: list, threshold: float, model_name: str):
    import torch
    import gc
    from transformers import AutoModelForSequenceClassification, AutoTokenizer

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForSequenceClassification.from_pretrained(model_name).to(device)
    model.eval()

    k, score_list = 0, []
    while k < len(docs):
        pairs = [[query, doc.page_content] for doc in docs[k : k + 10]]
        with torch.no_grad():
            inputs = tokenizer(
                pairs,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=512,
            ).to(device)
            scores = (
                model(**inputs, return_dict=True)
                .logits.view(
                    -1,
                )
                .float()
            )
        if k == 0:
            score_list = scores
        else:
            score_list = torch.cat([score_list, scores], dim=0)
        k += 10

    # Get the sorted indices in descending order
    sorted_indices = torch.argsort(score_list, descending=True)

    # Convert the indices to a Python list
    sorted_indices_list = sorted_indices.tolist()

    # Get the documents in the order of their scores, if lower than threshold, break
    documents = []
    for i in sorted_indices_list:
        if score_list[i] < threshold:
            break
        documents.append(docs[i])

    del model, inputs, tokenizer
    gc.collect()
    torch.cuda.empty_cache()

    return documents, score_list[: len(documents)]


def rerank_reduce(query, docs, topK):
    import torch
    import gc
    from transformers import AutoModelForSequenceClassification, AutoTokenizer

    model_name = "BAAI/bge-reranker-large"  # BAAI/bge-reranker-base
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForSequenceClassification.from_pretrained(model_name).to(device)
    model.eval()
    # topK //= 2
    k, score_list = 0, []
    while k < len(docs):
        pairs = [[query, doc.page_content] for doc in docs[k : k + 10]]
        with torch.no_grad():
            inputs = tokenizer(
                pairs,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=512,
            ).to(device)
            scores = (
                model(**inputs, return_dict=True)
                .logits.view(
                    -1,
                )
                .float()
            )
        if k == 0:
            score_list = scores
        else:
            score_list = torch.cat([score_list, scores], dim=0)
        k += 10

    # Get the sorted indices in descending order
    sorted_indices = torch.argsort(score_list, descending=True)

    # Convert the indices to a Python list
    sorted_indices_list = sorted_indices.tolist()

    # Get the documents in the order of their scores, if lower than threshold, break
    documents = []
    for i in sorted_indices_list:
        if i > topK:
            break
        documents.append(docs[i])

    del model, inputs, tokenizer
    gc.collect()
    torch.cuda.empty_cache()

    return documents



from typing import Any, List, Optional, Tuple
from pydantic import Field
import numpy as np
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from akasha.utils.db.db_structure import dbs
from langchain_community.retrievers import KNNRetriever


class myKNNRetriever(BaseRetriever):
    embeddings: Embeddings = Field(default=None)
    """Embeddings model to use."""
    index: Any = Field(default=None)
    """Index of embeddings."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict]
    k: int = 3
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None

    @classmethod
    def from_db(
        cls,
        db: dbs,
        embeddings: Embeddings,
        k: int = 3,
        relevancy_threshold: float = 0.0,
        **kwargs: Any,
    ) -> KNNRetriever:
        # db_data = _get_all_docs(db)
        index = np.array(db.get_embeds())
        texts = db.get_docs()
        metadata = db.get_metadatas()
        return cls(
            embeddings=embeddings,
            index=index,
            texts=texts,
            metadata=metadata,
            k=k,
            relevancy_threshold=relevancy_threshold,
            **kwargs,
        )

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement k-means search to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """
        query_embeds = np.array(self.embeddings.embed_query(query))
        # calc L2 norm
        index_embeds = self.index / np.sqrt((self.index**2).sum(1, keepdims=True))
        query_embeds = query_embeds / np.sqrt((query_embeds**2).sum())

        similarities = index_embeds.dot(query_embeds)
        sorted_ix = np.argsort(-similarities)

        denominator = np.max(similarities) - np.min(similarities) + 1e-6
        normalized_similarities = (similarities - np.min(similarities)) / denominator
        # print([normalized_similarities[row]
        #        for row in sorted_ix[0:self.k]])  # stats

        top_k_scores = [normalized_similarities[row] for row in sorted_ix[0 : self.k]]
        top_k_results = [
            Document(page_content=self.texts[row], metadata=self.metadata[row])
            for row in sorted_ix[0 : self.k]
            if (
                self.relevancy_threshold is None
                or normalized_similarities[row] >= self.relevancy_threshold
            )
        ]
        return top_k_results, top_k_scores

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    def _aget_relevant_documents(self, query: str) -> List[Document]:
        """implement k-means search to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """
        return self._gs(query)[0]

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]



from typing import List
from langchain.schema import Document
from dotenv import dotenv_values
import os
from akasha.utils.prompts.format import websearch_language_dict, websearch_country_dict


def load_docs_from_webengine(
    prompt: str,
    search_engine: str = "wiki",
    search_num: int = 5,
    language: str = "ch",
    env_file: str = "",
) -> List[Document]:
    """get the search results based on the prompt and search engine
    search_engine include : wiki, serper, brave, tavily

    """

    search_engine = search_engine.lower()

    if search_engine == "wiki":
        api_key = ""
    else:
        api_key = _get_search_api_key(search_engine, env_file)

    if search_engine == "wiki":
        from langchain_community.document_loaders import WikipediaLoader

        docs = WikipediaLoader(
            query=prompt,
            load_max_docs=search_num,
            lang=websearch_language_dict[language][search_engine],
        ).load()

    elif search_engine == "serper":
        from langchain_community.utilities import GoogleSerperAPIWrapper

        google_serper = GoogleSerperAPIWrapper(
            serper_api_key=api_key,
            gl=websearch_country_dict[language][search_engine],
            hl=websearch_language_dict[language][search_engine],
            k=search_num,
        )
        search_res = google_serper.run(prompt)

        docs = [Document(page_content=search_res)]

    elif search_engine == "brave":
        from langchain_community.document_loaders import BraveSearchLoader

        loader = BraveSearchLoader(
            query=prompt,
            api_key=api_key,
            search_kwargs={
                "count": search_num,
                "country": "all",
                "search_lang": websearch_language_dict[language][search_engine],
            },
        )
        docs = loader.load()

    elif search_engine == "tavily":
        from tavily import TavilyClient

        loader = TavilyClient(api_key=api_key)

        results = loader.search(prompt, "advanced", max_results=search_num)
        docs = [
            Document(page_content=doc_dict["title"] + "\n" + doc_dict["content"])
            for doc_dict in results["results"]
        ]

    else:
        raise ValueError(f"search_engine {search_engine} is not supported")
    return docs


def _get_search_api_key(search_engine: str = "wiki", env_file: str = "") -> str:
    """get the search api key based on the search engine"""
    if env_file == "" or not os.path.exists(env_file):
        if search_engine == "serper":
            return os.environ["SERPER_API_KEY"]
        elif search_engine == "brave":
            return os.environ["BRAVE_API_KEY"]
        elif search_engine == "tavily":
            return os.environ["TAVILY_API_KEY"]
        else:
            return ""

    else:
        env_dict = dotenv_values(env_file)

        if search_engine == "serper":
            return env_dict["SERPER_API_KEY"]
        elif search_engine == "brave":
            return env_dict["BRAVE_API_KEY"]
        elif search_engine == "tavily":
            return env_dict["TAVILY_API_KEY"]
        else:
            return ""



from typing import Any, List, Optional, Tuple, Dict, Iterable
from langchain_community.retrievers import TFIDFRetriever
from langchain.schema import Document
from langchain_core.callbacks import CallbackManagerForRetrieverRun
import jieba


class myTFIDFRetriever(TFIDFRetriever):
    @classmethod
    def from_texts(
        cls,
        texts: Iterable[str],
        documents: Optional[List[Document]] = None,
        tfidf_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> TFIDFRetriever:
        try:
            from sklearn.feature_extraction.text import TfidfVectorizer
        except ImportError:
            raise ImportError(
                "Could not import scikit-learn, please install with `pip install "
                "scikit-learn`."
            )

        tfidf_params = tfidf_params or {}
        vectorizer = TfidfVectorizer(**tfidf_params)
        tfidf_array = vectorizer.fit_transform(texts)
        return cls(
            vectorizer=vectorizer, docs=documents, tfidf_array=tfidf_array, **kwargs
        )

    @classmethod
    def from_documents(
        cls,
        documents: List[Document],
        *,
        tfidf_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> TFIDFRetriever:
        texts, metadatas = zip(
            *(
                (" ".join(list(jieba.cut(d.page_content))), d.metadata)
                for d in documents
            )
        )
        return cls.from_texts(
            texts=texts, tfidf_params=tfidf_params, documents=documents, **kwargs
        )

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun = None
    ) -> List[Document]:
        return self._gs(query)[0]

    def _gs(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        from sklearn.metrics.pairwise import cosine_similarity

        query_vec = self.vectorizer.transform(
            [query]
        )  # Ip -- (n_docs,x), Op -- (n_docs,n_Feats)
        results = cosine_similarity(self.tfidf_array, query_vec).reshape(
            (-1,)
        )  # Op -- (n_docs,1) -- Cosine Sim with each doc
        idxs = results.argsort()[-self.k :][::-1]
        return_docs = [self.docs[i] for i in idxs]
        return_values = [results[i] for i in idxs]

        return return_docs, return_values

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    async def _aget_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]
